# see notes http://www.ics.uci.edu/~pattis/ICS-21/lectures/notes/Skip%20Lists.pdf

@everywhere import Base.show
@everywhere import Base.isempty
@everywhere import Base.start
@everywhere import Base.done
@everywhere import Base.next
@everywhere import Base.length

@everywhere type Node
    key     ::  Number
    next    ::  Union(Node, Type{None})
    prev    ::  Union(Node, Type{None})
    below   ::  Union(Node, Type{None})
    above   ::  Union(Node, Type{None})

    Node(k::Number) = new(k, None, None, None, None)
end

@everywhere type SkipList
    s   ::  Node # s is the start position of the Skip List
    h   ::  Int  # h is the height of the Skip List
    n   ::  Int  # n is the number of items in the Skip List

    SkipList() = begin
        head = Node(-Inf)
        tail = Node(Inf)

        head.next = tail
        tail.prev = head

        new(head, 0, 0)
    end
end

@everywhere function search(skip::SkipList, k::Number)
    p = skip.s
    while p.below != None
        p = p.below
        while k >= p.next.key
            p = p.next
        end
    end
    p.key == k
end

@everywhere function search(skip::SkipList, a::Number, b::Number)
    p = skip.s
    while p.below != None
        p = p.below
        while a >= p.next.key
            p = p.next
        end
    end
    A = Number[]
    while p.key <= b
        if p.key >= a
            append!(A, [p.key])
        end
        p = p.next
    end
    A
end

## General Collections ################
@everywhere length(skip::SkipList) = skip.n
@everywhere isempty(skip::SkipList) = skip.n == 0
#######################################

## Iterable Collections ###############
@everywhere function start(skip::SkipList)
    s = skip.s
    while s.below != None
        s = s.below
    end
    s.next
end

@everywhere function done(skip::SkipList, node)
    node.key == Inf
end

@everywhere function next(skip::SkipList, node)
    node.key, node.next
end
#######################################


# insert the key k into the Skip List skip
@everywhere function insert(skip::SkipList, k::Number)
    # returns a node p in the bottom of the Skip List such that
    # p.key is the largest key <= k
    function SkipSearch(skip::SkipList, k::Number)
        p = skip.s
        while p.below != None
            p = p.below
            while k >= p.next.key
                p = p.next
            end
        end
        p
    end

    # insertAfterAbove inserts a node with key k after the node p and above
    # above the node q; returns the new node r
    function insertAfterAbove(p::Union(Node, Type{None}),
                              q::Union(Node, Type{None}), k::Number)
        r = Node(k)
        if p != None
            r.next = p.next
            if r.next != None
                r.next.prev = r
            end
            p.next = r
            r.prev = p
        end
        if q != None
            r.below = q
            q.above = r
        end
        r
    end

    p = SkipSearch(skip, k)
    q = None
    i = -1
    while true
        i += 1
        if i >= skip.h
            skip.h += 1
            t = skip.s.next
            skip.s = insertAfterAbove(None, skip.s, -Inf)
            insertAfterAbove(skip.s, t, Inf)
        end
        q = insertAfterAbove(p,q,k)
        while p.above == None
            p = p.prev
        end
        p = p.above

        if randbool()
            break
        end
    end
    skip.n += 1
    true
end

@everywhere function delete(skip::SkipList, k::Number)
    # returns a node p in the bottom of the Skip List such that
    # p.key is the largest key <= k
    function SkipSearch(skip::SkipList, k::Number)
        p = skip.s
        while p.below != None
            p = p.below
            while k >= p.next.key
                p = p.next
            end
        end
        p
    end

    q = SkipSearch(skip, k)
    if q.key != k
        error("$k is not in the skip list")
    end
    while q != None
        p = q.prev
        r = q.next
        p.next = r
        r.prev = p
        q = q.above
    end
    skip.n -= 1
    
    while skip.s.below.next.key == Inf
        skip.s = skip.s.below
    end
    true
end

@everywhere show(io::IO, node::Node) = print("Node(", node.key, ")")

@everywhere function show(io::IO, skip::SkipList)
    if skip.n <= 32
        p = skip.s
        while p != None
            t = p
            while t != None
                if t.next != None
                    print(t.key, " ->  ")
                else
                    print(t.key)
                end
                t = t.next
            end
            if p.below != None
                println("\n")
            end
            p = p.below
        end
    else
        print("SkipList with ", skip.n, " elements")
    end
end

@everywhere skipmap = Dict{Int, SkipList}()

@everywhere function createskip()
    n = 0
    while haskey(skipmap, n)
        n = int(rand()*10000000)
    end
    skipmap[n] = SkipList()
    myid(), n
end

type DSkipList
    skiplists

    DSkipList() = new(mapreduce(fetch, vcat, {@spawnat p createskip() for p=procs()}))
end

function insert(dskip::DSkipList, k::Number)
    sl = dskip.skiplists
    (rproc, sidx) = sl[rand(1:length(sl))]
    remotecall(rproc, ()->insert(skipmap[sidx], k))
end

function search(dskip::DSkipList, k::Number)
    fetch_or = (a,b) -> fetch(a)[1] || fetch(b)[1]
    A = {remotecall(p, ()->(search(skipmap[sidx], k), myid())) for (p,sidx) = dskip.skiplists}
    reduce(fetch_or, A)
end

function search(dskip::DSkipList, a::Number, b::Number)
    fetch_vcat = (a,b) -> vcat(fetch(a),fetch(b))
    A = {remotecall(p, ()-> search(skipmap[sidx], a, b)) for (p,sidx) = dskip.skiplists}
    reduce(fetch_vcat, A)
end

function delete(dskip::DSkipList, k::Number)
    sl = dskip.skiplists
    A = {remotecall(p, ()->(search(skipmap[sidx], k), myid())) for (p,sidx) = dskip.skiplists}
    A = filter((a) -> fetch(a)[1] == true, A)
    rproc = fetch(A[rand(1:length(A))])[2]
    sidx = filter((a) -> a[1] == rproc,sl)[1][2]
    remotecall(rproc, ()->delete(skipmap[sidx], k))
end
